/*
 * This is the DECtalk PC firmware loader for the Linux kernel, version 1.0
 *
 * Original 386BSD source:
 *      Copyright (c) 1996 Brian Buhrow <buhrow@lothlorien.nfbcal.org>
 *
 * Adapted for Linux:
 *      Copyright (c) 1997 Nicolas Pitre <nico@cam.org>
 *
 * Maintained by:
 *      Nicolas Pitre <nico@cam.org>
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/**************************************************************************

NAME: Brian Buhrow
DATE: April 20, 1996
MODIFIED BY: Nicolas Pitre
DATE: September 24, 1997

This program initializes and loads the Dectalk PC Speech synthesizer with
its firmware and dictionary information.  It relies on ioctls in the device
driver to actually accomplish the communication with the card itself.
However, it does the calculations to find the exe header offsets, etc.
this code is directly taken from DEC's DOS based driver.

**************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <sys/types.h>
#include <fcntl.h>
#include <malloc.h>
#include <sys/errno.h>

#include "dtpc_ioctl.h"
#include "dtload.h"

int verbose = 0;		/*verbosity level */
int dtfd;			/*speech device file descriptor */


void 
loadkernel (char *filname)
{
  FILE *kernfil;
  char *read_buff;		/*where do we put the kernel in memory to load it? */
  struct dt_kerninfo dt_data;	/*structure we pass to the driver for loading */
  int i, read_size;
  unsigned int total_paras;
  long image_left, image_done, fix;
  struct dos_exe_header header;
  struct dos_reloc *fixups;

  kernfil = fopen (filname, "r");
  if (!kernfil)
    {
      perror ("Opening kernel file: ");
      exit (2);
    }

  /*Get the buffer ready for the card's kernel */
  read_buff = (char *) malloc (READ_BUFF_SIZE);
  if (!read_buff)
    {
      fprintf (stderr, "Memory allocation failure for read_buff\n");
      exit (1);
    }

  /*get the fixups struct going */
  fixups = (struct dos_reloc *) malloc (sizeof (struct dos_reloc) * MAX_FIXUPS);
  if (!fixups)
    {
      fprintf (stderr, "Cannot allocate memory for fixups\n");
      exit (1);
    }

  /*
   *  Now, open up the .exe file and extract all the cgroup data and
   *  write that out to the .abs file ...
   */

  fread (&header, sizeof (struct dos_exe_header), 1, kernfil);
  if (header.id != 0x5a4d)
    {
      fprintf (stderr, "Invalid header file format\n");
      fprintf (stderr, "Want 0x5a4d, got 0x%x\n", header.id);
      exit (1);
    }
  if (header.relen > MAX_FIXUPS)
    {
      fprintf (stderr, "Too many fixups\n");
      exit (1);
    }

  fseek (kernfil, (long) header.reloc, SEEK_SET);
  fread (fixups, sizeof (struct dos_reloc) * header.relen, 1, kernfil);

/*
 * load in and fix up the image ...
 */

  fseek (kernfil, (long) (header.hsize * 16), SEEK_SET);
  image_left = ((long) (header.pages - 1) * 512) + (long) header.rem - ((long) header.hsize * 16);
  total_paras = (unsigned int) (image_left >> 4) + header.hmin + 16;
  image_done = 0;

  header.csval += 0x40;
  header.ssval += 0x40;

  if (verbose)
    {
      printf ("Loading kernel of %ld bytes (%d relocs)\n",
	      image_left, header.relen);
      printf ("    cs:ip == %04x:%04x   ss:sp == %04x:%04x\n",
	      header.csval, header.ipval, header.ssval, header.spval);
    }

  if (image_left > READ_BUFF_SIZE)
    {
      fprintf (stderr, "Image too large for buffer\n");
      exit (1);
    }
  read_size = (int) image_left;
  fread (read_buff, read_size, 1, kernfil);
  for (i = 0; i < header.relen; i++)
    {
      fix = (fixups[i].segment << 4) + fixups[i].offset;
      (*(unsigned int *) &read_buff[fix]) += 0x40;
    }

/*
 *  load up the image ...
 */

  /*initialize our dt_data structure */
  bzero (&dt_data, sizeof (struct dt_kerninfo));
  dt_data.dt_start = read_buff;
  dt_data.dt_len = read_size;
  dt_data.dt_csval = header.csval;
  dt_data.dt_ipval = header.ipval;

  /*actually go for the load */
  if (ioctl (dtfd, DTIOLDKERNEL, &dt_data) < 0)
    {
      perror ("Error loading kernel into the card: ");
      exit (1);
    }

  fclose (kernfil);
  free (read_buff);
  free (fixups);
}



/*load a specified dictionary */
void 
loaddict (char *filnam, char *name, int type)
{
  FILE *fdic;			/*dictionary file pointer */
  char *read_buff;
  int i, read_index, read_size, act_size;
  unsigned short *index_fix, seg_fix;
  unsigned long entries, index, dic_bytes, load_addr, dic_addr, total_done;
  unsigned int total_paras;
  struct dt_loadinfo dt_loadreq;	/* Use this to load  memory */
  struct dt_dicinfo dt_dicreq; 	/* Registered loaded dictionaries with this */
  unsigned long param;

  fdic = fopen (filnam, "r");
  if (!fdic)
    {
      perror (filnam);
      exit (1);
    }

  /*Allocate a buffer for us to write in */
  read_buff = (char *) malloc (READ_BUFF_SIZE);
  if (!read_buff)
    {
      fprintf (stderr, "Error allocating read buffer\n");
      exit (1);
    }

/*
 * read in the entry count and the actual entry size excluding the
 * index table (which is entries * 4) ...
 */

  fread (&entries, 4, 1, fdic);
  fread (&dic_bytes, 4, 1, fdic);
  if (verbose)
    printf ("Loading %s dictionary of %lu entries, %lu bytes.\n",
	    name, entries, dic_bytes);

  total_paras = (((entries * 4) + dic_bytes) >> 4) + 2;
  if (verbose)
    printf ("Allocating %d paragraphs of free ram ...\n", total_paras);

  param = total_paras;
  if (ioctl (dtfd, DTIOALLOCMEM, &param) < 0)
    {
      perror ("Error requesting memory from speech device");
      exit (1);
    }
  load_addr = param;

  seg_fix = (load_addr >> 4) & 0xffff;
  dic_addr = load_addr;

/*
 *  read, reloc and load a bunch of indexes ...
 */

  index = entries;
  total_done = 0;
  index_fix = (unsigned short *) &read_buff[0];
  if (verbose)
    printf ("Index table starts at %lx\n", load_addr);
  while (index > 0)
    {
      read_index = (index > (READ_BUFF_SIZE / 4)) ? (READ_BUFF_SIZE / 4) : index;
      if ((act_size = fread (read_buff, 1, read_index * 4, fdic)) != read_index * 4)
	{
	  fprintf (stderr, "\nError reading indexes\n");
	  fprintf (stderr, "    exp : %d  act : %d\n", read_index * 4, act_size);
	  exit (1);
	}
      for (i = 1; i < read_index * 2; i += 2)
	index_fix[i] += seg_fix;
      dt_loadreq.dtl_start = read_buff;
      dt_loadreq.dtl_len = read_index*4;
      dt_loadreq.dtl_addr = load_addr;
      if (ioctl (dtfd, DTIOLDMEM, &dt_loadreq) < 0)
	{
	  fprintf (stderr, "\nError loading indexes at 0x%lX: ", load_addr);
	  perror ("");
	  exit (1);
	}
      load_addr += read_index * 4;
      index -= read_index;
      total_done += read_index;
      if (verbose)
	printf ("\rLoading indexes ... %ld", total_done);
    }

/*
 *  now, load up the dictionary bytes ...
 */
  if (verbose)
    printf ("\nDictionary text starts at %lx\n", load_addr);
  total_done = 0;
  while (dic_bytes > 0)
    {
      read_size = (dic_bytes > READ_BUFF_SIZE) ? READ_BUFF_SIZE : (int) dic_bytes;
      if ((act_size = fread (read_buff, 1, read_size, fdic)) != read_size)
	{
	  fprintf (stderr, "\nError reading dictionary text!\n");
	  fprintf (stderr, "    asked : %d  actual : %d\n", act_size, read_size);
	  exit (1);
	}
      dt_loadreq.dtl_start = read_buff;
      dt_loadreq.dtl_len = read_size;
      dt_loadreq.dtl_addr = load_addr;
      if (ioctl (dtfd, DTIOLDMEM, &dt_loadreq) < 0)
	{
	  fprintf (stderr, "\nError loading dictionary at 0x%lX: ", load_addr);
	  perror ("");
	  exit (1);
	}
      dic_bytes -= read_size;
      load_addr += read_size;
      total_done += read_size;
      if (verbose)
	printf ("\rLoading text ... %ld", total_done);
    }
  dt_dicreq.dtd_type = type;
  dt_dicreq.dtd_addr = dic_addr;
  dt_dicreq.dtd_entries = entries;
  if (ioctl (dtfd, DTIOSETDICT, &dt_dicreq) < 0)
    {
      fprintf (stderr, "\nError setting up dictionary on speech device: ");
      perror ("");
      exit (1);
    }
  if (verbose)
    printf("\n");

  fclose (fdic);
  free (read_buff);
}


/*Load and start executables */
void 
loadexe (char *filnam)
{
  FILE *fexe;
  char *read_buff;
  unsigned int load_addr = 0;
  struct dos_reloc *fixups;

  int i, read_size;
  unsigned int total_paras, seg_fix;
  long image_left, image_done, fix;
  struct dos_exe_header header;
  struct dt_loadinfo dt_loadreq;	/*load request info */
  struct dt_taskinfo dt_taskreq;	/*task request info */
  unsigned long param;

  if (verbose)
    printf ("Loading %s as an exec file\n", filnam);
  /*open the file, and make sure we can read it */
  fexe = fopen (filnam, "r");
  if (!fexe)
    {
      perror (filnam);
      exit (1);
    }

  /*Make sure we have enough memory */
  read_buff = (char *) malloc (READ_BUFF_SIZE);
  if (!read_buff)
    {
      fprintf (stderr, "Unable to allocate memory for load buffer: ");
      perror ("");
      exit (1);
    }

  /*allocate memory for the dos relocation structure */
  fixups = (struct dos_reloc *) malloc (sizeof (struct dos_reloc) * MAX_FIXUPS);
  if (!fixups)
    {
      fprintf (stderr, "Unable to allocate relocation buffers: ");
      perror ("");
      exit (1);
    }

/*
 *  Now, open up the .exe file and extract all the cgroup data and
 *  write that out to the .abs file ...
 */

  fread (&header, sizeof (struct dos_exe_header), 1, fexe);
  if (header.id != 0x5a4d)
    {
      fprintf (stderr, "Invalid .exe file format -- no signature in file %s\n",
	       filnam);
      return;
    }

  if (header.relen <= MAX_FIXUPS)
    {
      fseek (fexe, (long) header.reloc, SEEK_SET);
      fread (fixups, sizeof (struct dos_reloc) * header.relen, 1, fexe);
    }
  else
    {
      fprintf (stderr, "Can't load image ... too many fixups\n");
      return;
    }

/*
 * load in and fix up the image ...
 */

  fseek (fexe, (long) (header.hsize * 16), SEEK_SET);
  image_left = ((long) (header.pages - 1) * 512) + (long) header.rem - ((long) header.hsize * 16);
  total_paras = (unsigned int) (image_left >> 4) + header.hmin + 16;
  image_done = 0;

  if (load_addr == 0)
    {
      param = total_paras;
      if (ioctl (dtfd, DTIOALLOCMEM, &param) < 0)
	{
	  fprintf (stderr, "Error allocating memory on card: ");
	  perror ("");
	  exit (1);
	}
      load_addr = param;
      if (!load_addr)
	{
	  fprintf (stderr, "Error allocating memory on module\n");
	  return;
	}
    }
  seg_fix = (load_addr >> 4) & 0xffff;

  if (verbose)
    {
      printf ("Loading image of %ld bytes (%d relocs)\n",
	      image_left, header.relen);
      printf ("Allocating %ld bytes of free ram at %05x\n",
	      (long) header.hmin * 16, load_addr);
      printf ("Total memory taken is %ld bytes\n", (long) total_paras * 16);
      printf ("    cs:ip == %04x:%04x   ss:sp == %04x:%04x\n",
	      header.csval + seg_fix, header.ipval, header.ssval + seg_fix, header.spval);
    }

  while (image_left > 0)
    {
      read_size = (image_left > (long) READ_BUFF_SIZE) ? READ_BUFF_SIZE : (int) image_left;
      fread (read_buff, read_size, 1, fexe);
      for (i = 0; i < header.relen; i++)
	{
	  fix = ((long) fixups[i].segment << 4) + (long) fixups[i].offset;
	  if (fix >= image_done && fix < image_done + read_size)
	    (*(unsigned int *) &read_buff[fix - image_done]) += seg_fix;
	}
      dt_loadreq.dtl_start = read_buff;
      dt_loadreq.dtl_len = read_size;
      dt_loadreq.dtl_addr = load_addr + image_done;
      if (ioctl (dtfd, DTIOLDMEM, &dt_loadreq) < 0)
	{
	  fprintf (stderr, "Error loading speech device at 0x%lX: ", load_addr + image_done);
	  perror ("");
	  exit (1);
	}
      image_left -= read_size;
      image_done += read_size;
    }
  dt_taskreq.dts_csval = header.csval + seg_fix;
  dt_taskreq.dts_ipval = header.ipval;
  if (ioctl (dtfd, DTIOSTARTEXE, &dt_taskreq) < 0)
    {
      fprintf (stderr, "Error starting %s: ", filnam);
      perror ("");
      exit (1);
    }

  fclose (fexe);
  free (read_buff);
  free (fixups);
}



/*usage routine */
void 
usage (char *argv)
{
  fprintf (stderr, "Usage: %s -f <speech device> -K <kernel name> "
	   "-D <primary dict> [-vtx] <secondary dict> <executable> "
	   "...\n", argv);
}


int
main (argc, argv)
     int argc;
     char **argv;
{
  int c, i, verbalinit, skipinst;
  int needkernel, needdict;
  char *kernfil, *dicfil, *dtkfil, *cp;

  /* process command line arguments and set up defaults */
  needkernel = needdict = 0;
  kernfil = dicfil = dtkfil = NULL;	/*no clue, as yet */
  verbalinit = skipinst = 0;

  while ((c = getopt (argc, argv, ":K:f:D:vtx")) != -1)
    {
      switch (c)
	{
	case 'K':
	  kernfil = optarg;
	  needkernel = 1;
	  break;
	case 'f':
	  dtkfil = optarg;
	  break;
	case 'D':
	  dicfil = optarg;
	  needdict = 1;
	  break;
	case 'v':
	  verbose = 1;
	  break;
	case 't':
	  verbalinit = 1;
	  break;
	case 'x':
	  skipinst = 1;		/*if we're loaded, skip all inits */
	  break;
	case ':':
	  fprintf (stderr, "%s: option %c requires an argument\n", argv[0], optopt);
	  break;
	default:
	  usage (argv[0]);
	  exit (1);
	  break;
	}
    }

  /*make sure we have a dtkfil for use.  We'll also try the environment */
  if (!dtkfil)
    {
      dtkfil = getenv ("DTK_PORT");
      if (!dtkfil)
	{
	  fprintf (stderr, "%s: Unable to determine speech device\n", argv[0]);
	  fprintf (stderr, "Try checking your usage or setting the DTK_PORT environment variable\n");
	  exit (1);
	}
    }

  /*open the speech device and determine what state it's in */
  dtfd = open (dtkfil, O_RDWR, 0666);
  if (dtfd < 0)
    {
      fprintf (stderr, "%s: Unable to open %s: ", argv[0], dtkfil);
      perror ("");
      exit (1);
    }
  if ((ioctl (dtfd, DTIOWARM, NULL)) && (errno == EINVAL))
    {
      needkernel = needdict = 0;
    }

  /*If we need  a kernel, and we don't have a file, bail.  Otherwise, load */
  if (needkernel)
    {
      if (!kernfil)
	{
	  fprintf (stderr, "%s: Unable to load kernel for %s\n", argv[0], dtkfil);
	  exit (1);
	}
      loadkernel (kernfil);
    }

  /*If we need the primary dictionary, but can't findit, talk. */
  if (needdict)
    {
      if (!dicfil)
	{
	  fprintf (stderr, "%s: Unable to load premary dictionary for %s\n",
		   argv[0], dtkfil);
	  exit (1);
	}
      loaddict (dicfil, "primary", PRIMARY_DIC);
    }

  /* If skipinst is set, it means the user doesn't want to load any
   * secondary dictionaries if the kernel is already loaded.
   */
  if ((skipinst) && (!needkernel))
    return 0;

  /*process secondary dictionaries and programs */
  for (i = optind; i < argc; i++)
    {
      cp = index (argv[i], '.');
      if (!cp)
	loadexe (argv[i]);
      cp++;			/*get beyond the "." in the file name */
      if (!strncmp (cp, "dtu", 3))
	{
	  loaddict (argv[i], "user", USER_DIC);
	  continue;
	}
      if (!strncmp (cp, "dta", 3))
	{
	  loaddict (argv[i], "abbreviation file", ABBREV_DIC);
	  continue;
	}
      loadexe (argv[i]);
    }
  /*If verbal confirmation of the initialization is desired, and we
     loaded the kernel for this run, do it here. */

  if ((needkernel) && ((verbose) || (verbalinit)))
    ioctl (dtfd, DTIOTEST, NULL);
  close (dtfd);
  return 0;
}
