/*
 * This is the DECtalk PC driver for the Linux kernel, version 0.95
 * 
 * Original 386BSD source: 
 * 	Copyright (c) 1996 Brian Buhrow <buhrow@lothlorien.nfbcal.org>
 *
 * Rewritten for Linux:
 *  	Copyright (c) 1997 Nicolas Pitre <nico@cam.org>
 *
 * ChangeLog:
 * - Feb 22, 1999:
 *	Minor compilation fix for glibc
 *
 * - Apr 15, 1999:
 *	Ported to Linux 2.2.x by Jim Danley <Jim@danley.com>
 * 
 * - May 02, 1999:
 *	Added compatibility macros so it compiles either for Linux 2.0.x
 *	or Linux 2.2.x (Nico)
 *
 * Maintained by: 
 *	Nicolas Pitre <nico@cam.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/timer.h>
#include <linux/termios.h>

#include <asm/system.h>
#include <asm/io.h>

#include "dtpc_reg.h"
#include "dtpc_ioctl.h"



// #define LOG_DATA	/* send back incoming data on minor LOG_MINOR */
// #define DEBUG	/* prints or fill up syslog with debugging messages */



static const char Hello[] = {
  "DECtalk PC driver for Linux v0.95, (C) 1997-1999 by Nicolas Pitre <nico@cam.org> \n"
};



#include <linux/version.h>
#if LINUX_VERSION_CODE < 0x000200ff
# define KERNEL_2_0_x
#endif

#ifdef KERNEL_2_0_x
/* compatibility macros for Linux 2.0.x compilation */
# define DEVICE_ID  (inode->i_rdev)
# define signal_pending( current )  ((current)->signal & ~((current)->blocked))
# define schedule_timeout( x )  {current->timeout = jiffies + (x); schedule();}
# define copy_from_user  memcpy_fromfs
# define copy_to_user  memcpy_tofs
# define Get_user( x, y )  (x) = get_user( (y) )
#else
/* compatibility macros for Linux 2.2.x compilation */
# define DEVICE_ID  (file->f_dentry->d_inode->i_rdev)
# include <asm/uaccess.h>
# include <linux/poll.h>
# define Get_user  get_user
#endif



#ifdef DEBUG
  #define dprintk( a, arg... ) printk( a, ##arg )
#else
  #define dprintk( a, arg... )
#endif



#ifdef LOG_DATA
  #define DT_LOG_MINOR 16
  #define LOG_BUF_SZ 8192
  static char log_buf[LOG_BUF_SZ];
  static int volatile log_buf_in = 0;
  static int volatile log_buf_out = 0;
  static int volatile log_buf_tot = 0;
  static struct wait_queue *log_wait = NULL;
#endif



/* Number of simultaneous boards */
#define NR_CARDS 4


/* Possible I/O base addresses.  It is possible to change them to fit your 
 * hardware if you have a DECtalk PC2 that is set to another port.  
 * Those listed here are the possible values for the DECtalkPC. 
 */
static u_short base[NR_CARDS] = { 0x250, 0x240, 0x350, 0x340 };


/* structure to keep info on each card */
struct dt_card {
  u_short short iobase;			/* base I/O adress */
  int status;
  int volatile flags;
  int volatile dt_dma_state;		/* toggle, last DECtalk dma flag */
  int volatile in_use;			/* = 1 if someone else is accessing the board */
  struct wait_queue *wait_read ;	/* waiting process for reads */
  struct wait_queue *wait_write ;	/* waiting process for writes */
  struct timer_list polling_timer;
  unsigned long polling_timeout;
  struct termios tios ;			/* termios for tty emulation */
};
static struct dt_card cards[NR_CARDS];


#define DT_HAS_KERNEL 0x0001 /* Does the board has its software kernel installed */
#define DT_IS_RUNNING 0x0002 /* is the board actually running that kernel? */
#define DT_ISOPEN     0x0004 /* Is the device opened by another process? */
#define DT_POLLING    0x0008 /* Is the driver running in polling mode */
#define DT_HASDICT    0x0010 /* does the card has its primary dictionary loaded */
#define DT_FLUSHED    0x0020 /* received a flush command */



#ifndef DT_MAJOR
  #define DT_MAJOR 0 	/* dynamic major number allocation if not defined */
#endif
static int dt_major = DT_MAJOR;



#define MIN(a,b)   ((a<b)?a:b)



/* 
 * Delay function to sleep when the card need some time...
 */

static void dt_delay( long ms )
{
  current->state = TASK_INTERRUPTIBLE;
  schedule_timeout((ms * HZ + 1000 - HZ) / 1000);
  current->state = TASK_RUNNING;
}



/* General purpose routines to query the card for status, ship comands,
 * etc.
 */

/* read status information from the indicated module */
static inline int dt_getstatus( u_short iobase )
{
  return( inb_p( iobase )|(inb_p( iobase+1 )<<8) );
}


/* send a command to the indicated module */
static inline void dt_sendcmd( u_short iobase, u_int cmd )
{
  outb_p( cmd & 0xFF, iobase );
  outb_p( (cmd>>8) & 0xFF, iobase+1 );
}


/* see if a certain bit is set */
static int dt_waitbit( u_short iobase, int bit )
{
  unsigned long timeout = jiffies + 5*HZ;  /* 5 sec */
  do{
    if( (dt_getstatus( iobase ) & bit ) == bit ) return( 1 );
    schedule();
  }while( jiffies < timeout );
  printk( "dt_waitbit: failed: status = 0x%X, bit = 0x%X\n",
          dt_getstatus( iobase ), bit );
  return 0;
}


/* Look for the given mode pattern */
static int dt_waitmode( u_short iobase, int pattern )
{
  unsigned long timeout = jiffies + 10*HZ;  /* 10 sec */
  do{
    if( dt_getstatus( iobase ) == pattern ) return( 1 );
    schedule();
  }while( jiffies < timeout );
  printk( "dt_waitmode: failed: status = 0x%X, pattern = 0x%X\n",
          dt_getstatus( iobase ), pattern );
  return( 0 );
}


/* This function waits for the dma state toggle to change on the card. */
static int dt_wait_dma( struct dt_card *card )
{
  unsigned long timeout = jiffies + 1 * HZ;
  u_short iobase = card->iobase;
  int state = card->dt_dma_state;

  /* Note: DMA toggle seems not to be active if the kernel is not loaded.
   * If so, we assume dma access is OK after a delay.
   */
  if( ! card->flags & DT_HAS_KERNEL ){
    dt_delay( 500 );
    return( dt_waitbit( iobase, STAT_dma_ready ) );
  }

  if( ! dt_waitbit( iobase, STAT_dma_ready ) ) return 0;
  do{
    if( (dt_getstatus(iobase)&STAT_dma_state) == state ) return 1;
    schedule();
  }while( jiffies < timeout );

  /* DMA toggle state time out */
  printk( KERN_DEBUG "dtpc: dma toggle state timed out: status = 0x%X, state = 0x%X\n",
          dt_getstatus( iobase ), state );
  /* update to current dma state */
  card->dt_dma_state = dt_getstatus( iobase ) & STAT_dma_state;
  return 1;
}
  

/* Ensure exclusive access to the card.
 * If returned value is non-zero, a signal has been received.
 * A card->in_use-- instruction must be made to mark the end of the 
 * exclusive access.
 */
static int dt_excl( struct dt_card *card )
{
  while( card->in_use ){
    schedule();
    if( signal_pending( current ) ) return -EINTR;
  }
  card->in_use++;
  return 0;
}



/* Sends a control command
 */
static int dt_ctrl( u_short iobase, u_int cmd, int val )
{
  if( ! dt_waitbit( iobase, STAT_cmd_ready ) ) return -EIO;
  outb_p( val&0xFF, iobase+2 );
  outb_p( (val>>8)&0xFF, iobase+3 );
  dt_getstatus( iobase );
  dt_sendcmd( iobase, CMD_control| cmd );
  outb_p( 0, iobase+6 );
  while( dt_getstatus( iobase ) & STAT_cmd_ready ) schedule();
  dt_sendcmd( iobase, CMD_null );
  return 0;
}



/* dt_flush()
 * This flushes any speech and all buffers on the card.
 */

static int dt_flush( struct dt_card *card )
{
  u_short iobase = card->iobase;
  int res;

  dprintk( "dt_flush, status = 0x%X\n", dt_getstatus( iobase ) );
  res = dt_ctrl( iobase, CTRL_flush, 0 );
  if( res ) return res;
  dt_waitbit( iobase, STAT_dma_ready );
  outb_p( DMA_sync, iobase+4 );
  outb_p( 0, iobase+4 );
  card->dt_dma_state ^= STAT_dma_state;
  while( dt_getstatus( iobase ) & STAT_flushing ) schedule();;
  card->flags |= DT_FLUSHED;
  return 0;
}


/* dt_flush_char()
 * Flush speech and say character.
 */

static int dt_flush_char( struct dt_card *card, unsigned char c )
{
  u_short iobase = card->iobase;

  dprintk( "dt_flush_char(%c), status = 0x%X\n", c, dt_getstatus( iobase ) );
  if( ! dt_waitbit( iobase, STAT_cmd_ready ) ) return -EIO;
  outb_p( 0, iobase+2 );
  outb_p( 0, iobase+3 );
  dt_getstatus( iobase );
  dt_sendcmd( iobase, CMD_control|CTRL_flush );
  outb_p( 0, iobase+6 );
  while( dt_getstatus( iobase ) & STAT_cmd_ready ) schedule();
  dt_sendcmd( iobase, CMD_null );
  dt_waitbit( iobase, STAT_dma_ready );
  outb_p( DMA_sync_char, iobase+4 );
  outb_p( c, iobase+4 );
  card->dt_dma_state ^= STAT_dma_state;
  while( dt_getstatus( iobase ) & STAT_flushing ) schedule();;
  return 0;
}


/* dt_sendbuff()
 * Takes a buffer pointer and a size parameter and writes that buffer to
 * the card.  The actual number of characters sent is returned.  If an error
 * occurs, -1 is returned.
 */
static int dt_sendbuff( struct dt_card *card, char *src, int len )
{
  u_short iobase = card->iobase;
  char *src0 = src;

  if( !len ) return 0;

#ifdef LOG_DATA
 { 
  /* We copy what we send in the log buffer */
  int i;
  for( i = 0; i < len; i++ ){
    log_buf[log_buf_in++] = src[i];
    log_buf_in %= LOG_BUF_SZ;
    log_buf_tot++;
    if( log_buf_tot >= LOG_BUF_SZ ) { 
      log_buf_out++; 
      log_buf_out %= LOG_BUF_SZ;
      log_buf_tot--;
    }
  }
  wake_up_interruptible( &log_wait );
 }
#endif

  /* send the buffer to the card */
  while( len-- ){
    if( ! dt_wait_dma( card ) ) return -1;
    if( ! (dt_getstatus( iobase ) & STAT_rr_char) ) break;
    outb_p( DMA_single_in, iobase+4 );
    outb_p( *src++, iobase+4 );
    card->dt_dma_state ^= STAT_dma_state;
  }
  dprintk( "dt_sendbuff: sent %d bytes\n", src - src0 );
  return( src - src0 );
}


/* dt_loadmem()
 * Takes the iobase address of the card in question, an address on that
 * card, and a pointer to a buffer containing the data to load, and loads
 * the data into the card at that address.
 * It returns 0 on success, -1 on failure.
 */
static int dt_loadmem( struct dt_card *card, u_int addr, u_int len, char *src )
{
  u_short iobase = card->iobase;
  char c;

  dprintk( "dt_loadmem: addr = %08X size = %d\n", addr, len );
  if( ! dt_wait_dma( card ) ) return -1;
  outb_p( DMA_control, iobase+4 );
  outb_p( DT_LOAD_MEM, iobase+4 );
  card->dt_dma_state ^= STAT_dma_state;

  if( ! dt_wait_dma( card ) ) return -1;
  outb_p( addr & 0xFF, iobase+4 );
  outb_p( (addr>>8) & 0xFF, iobase+4 );
  outb_p( (addr>>16) & 0xFF, iobase+4 );
  outb_p( (addr>>24) & 0xFF, iobase+4 );
  outb_p( len & 0xFF, iobase+4 );
  outb_p( (len>>8) & 0xFF, iobase+4 );
  card->dt_dma_state ^= STAT_dma_state;

  if( ! dt_wait_dma( card ) ) return -1;
  while( len-- ){ 
    Get_user( c, src++ );
    outb_p( c, iobase+4 );
    if( (len % 1024) == 0 ) schedule();  /* give some CPU to others */
  }
  card->dt_dma_state ^= STAT_dma_state;

  return( 0 );
}



/* 
 * This is the polling check to see if the card is ready to send/receive data.
 * The check is made through a timer task.  It gets the card's status and
 * wakes up read or write queue as needed.  If the timeout expires, both 
 * queues are woken up.  In this case, the poll timer is restarted by the
 * reading and/or writing process if it still exists.
 */

static void dt_polling_task( unsigned long l )
{
  struct dt_card *card = (struct dt_card *)l;
  int status = dt_getstatus( card->iobase );

  if( jiffies > card->polling_timeout ){
    /* timeout: force wake up */
    status |= STAT_tr_char|STAT_rr_char;
  }else if( card->flags & DT_FLUSHED ){
    /* a CTRL_a is available for read */
    status |= STAT_tr_char;
  }

  if( status & STAT_tr_char ) wake_up_interruptible( &card->wait_read );
  if( status & STAT_rr_char ) wake_up_interruptible( &card->wait_write );
  if( (status & STAT_tr_char) && (status & STAT_rr_char) ) return;

  /* prepare for another poll in 0.1 sec */
  card->polling_timer.expires = jiffies + HZ/10;
  add_timer( &card->polling_timer );
}


static void dt_polling( struct dt_card *card )
{
  /* we start the polling timer */
  del_timer( &card->polling_timer );
  card->polling_timer.function = dt_polling_task;
  card->polling_timer.data = (unsigned long)card;
  card->polling_timer.expires = jiffies + HZ/10;
  card->polling_timeout = jiffies + 5*HZ;	/* 5 sec */
  add_timer( &card->polling_timer );
  dprintk( "dt_start_polling\n" );
}



/* 
 * Kernel interface functions
 */

#ifdef KERNEL_2_0_x
static int dt_write( struct inode *inode,
		     struct file *file,
		     const char *src,
		     int count )
#else
static ssize_t dt_write( struct file *file, 
			 const char *src, 
			 size_t count,
			 loff_t *ppos )
#endif
{
  struct dt_card *card = &cards[MINOR(DEVICE_ID)];
  int i, remain;
  char c, buffer[256];

  dprintk( "dt_write: minor %d, flags %d, count %d\n", 
           MINOR(DEVICE_ID), file->f_flags, count );
  if( (card->flags & (DT_HAS_KERNEL|DT_IS_RUNNING|DT_HASDICT)) != 
      (DT_HAS_KERNEL|DT_IS_RUNNING|DT_HASDICT) ) return( -ENOSYS );

  /* check for ^C == async flush */
  src += count;
  remain = 0;
  while( remain < count ){
    Get_user( c, --src );
    if( c == 3 ) {
      src++;
      if( dt_excl( card ) ) return -ERESTARTNOINTR;
      dt_flush( card );
      card->in_use--;
      break;
    }
    remain++;
  }

  /* send data to the card */
  while( remain ){
    i = MIN( remain, 256 );
    copy_from_user( buffer, src, i );
    if( dt_excl( card ) ) break;;
    i = dt_sendbuff( card, buffer, i );
    card->in_use--;
    if( i < 0 ) return -EIO;
    if( i > 0 ){
      src += i;
      remain -= i;
      if( remain ) schedule();  /* give some CPU to others */
    }else{
      /* the card is not ready to receive data anymore */
      /* Should exit if non-blocking flag is on */
      if( (file->f_flags & O_NONBLOCK) == O_NONBLOCK ) break;
      /* let's sleep until the polling task wakes us up. */
      dt_polling( card );
      interruptible_sleep_on( &card->wait_write );
      if( signal_pending( current ) ) break;;
    }
  }

  return( count - remain );
}


#ifdef KERNEL_2_0_x
static int dt_read( struct inode *inode,
		    struct file *file,
		    char *dst,
		    int count )
#else
static ssize_t dt_read( struct file *file, 
			char *dst, 
			size_t count,
			loff_t *ppos )
#endif
{
  struct dt_card *card;
  u_short iobase;
  char *dst0 = dst;
  u_int i;

#ifdef LOG_DATA
  if( MINOR(DEVICE_ID) == DT_LOG_MINOR ){
    /* We return what's in the log buffer */
    while( count ){
      i = MIN( count, log_buf_tot );
      count -= i; 
      while( i-- ){
        put_user( log_buf[log_buf_out++], dst++ );
        log_buf_out %= LOG_BUF_SZ;
        log_buf_tot--;
      }
      if( dst != dst0 ) break;	/* ... so we return what we got up to now */
      if( (file->f_flags & O_NONBLOCK) == O_NONBLOCK ) break;
      interruptible_sleep_on( &log_wait );
      if( signal_pending( current ) ) return -ERESTARTSYS;
    }
    return( dst - dst0 );
  }
#endif

  card = &cards[MINOR(DEVICE_ID)];
  iobase = card->iobase;

  dprintk( "dt_read: minor %d, flags %d, count %d\n", 
           MINOR(DEVICE_ID), file->f_flags, count );
  if( (card->flags & (DT_HAS_KERNEL|DT_IS_RUNNING|DT_HASDICT)) != 
      (DT_HAS_KERNEL|DT_IS_RUNNING|DT_HASDICT) ) return( -ENOSYS );
 
  while( count ){
    /* simulate DECtalk Express return codes */
    if( card->flags & DT_FLUSHED ){
      card->flags &= ~DT_FLUSHED;
      put_user( 1, dst++ );	/* sends CTRL-A */
    }

    /* check if data is available */
    if( ! dt_wait_dma( card ) ) return -EIO;
    if( ! (dt_getstatus( iobase ) & STAT_tr_char ) ){
      /* no more data to read... */
      if( dst != dst0 ) break;	/* ... so we return what we got up to now */
      /* Should exit if non-blocking flag is on */
      if( (file->f_flags & O_NONBLOCK) == O_NONBLOCK ) break;
      /* let's sleep until the polling task wakes us up. */
      dt_polling( card );
      interruptible_sleep_on( &card->wait_read );
      if( signal_pending( current ) ) return -ERESTARTNOINTR;
    }

    /* Ensure exclusive access to the card */
    if( dt_excl( card ) ) return -ERESTARTNOINTR;

    /* query the card for incoming bytes */
    if( ! dt_wait_dma( card ) ){
      card->in_use--;
      return -EIO;
    }
    outb_p( DMA_buff_out, iobase+4 );
    outb_p( MIN( count, 128 ), iobase+4 ); /* request for bytes */
    card->dt_dma_state ^= STAT_dma_state;

    /* get the actual read count */
    if( ! dt_wait_dma( card ) ){
      card->in_use--;
      return -EIO;
    }
    i = inb_p( iobase+4 ) & 0xFF;
    dprintk( "dt_read: getting %d bytes\n", i );
    if( i > count ){  /* safety check... */
      printk( "dtpc: read: bad input count: %d\n", i );
      i = count;
    }
    /* copy data to input buffer */
    count -= i;
    while( i-- ) put_user( inb_p( iobase+4 ), dst++ );
    card->dt_dma_state ^= STAT_dma_state;
    card->in_use--;
  }

  return( dst - dst0 );
}


#ifdef KERNEL_2_0_x

static int dt_select( struct inode *inode, struct file *file,
		      int sel_type, select_table *wait )
{
  struct dt_card *card = &cards[MINOR(DEVICE_ID)];

  dprintk( "dt_select: type %d.\n", sel_type );

  if( (card->flags & (DT_HAS_KERNEL|DT_IS_RUNNING|DT_HASDICT)) !=
      (DT_HAS_KERNEL|DT_IS_RUNNING|DT_HASDICT) ) return( -ENOSYS );

  switch( sel_type ){
  case SEL_IN:
    if( card->flags & DT_FLUSHED   /* should return at least a CTRL-A to read */
        || dt_getstatus( card->iobase ) & STAT_tr_char ) break;
    dt_polling( card );
    select_wait( &card->wait_read, wait );
    return 0;
  case SEL_OUT:
    if( dt_getstatus( card->iobase ) & STAT_rr_char ) break;
    dt_polling( card );
    select_wait( &card->wait_write, wait );
    return 0;
  }
  dprintk( "dt_select: data is available\n" );
  return 1;
}

#else

static unsigned int dt_poll( struct file *file, poll_table *wait )
{
  struct dt_card *card = &cards[MINOR(DEVICE_ID)];
  u_int mask = 0;

  if( (card->flags & (DT_HAS_KERNEL|DT_IS_RUNNING|DT_HASDICT)) != 
      (DT_HAS_KERNEL|DT_IS_RUNNING|DT_HASDICT) ) return( -ENOSYS );

  /* should return at least a CTRL-A to read */
  if( card->flags & DT_FLUSHED || dt_getstatus( card->iobase ) & STAT_tr_char )
      mask |= POLLIN | POLLRDNORM;  /* readable */
  else {
      dt_polling( card );
      poll_wait( file, &card->wait_read, wait );
  }

  if( dt_getstatus( card->iobase ) & STAT_rr_char )
      mask |= POLLOUT | POLLWRNORM; /* writable */
  else {
      dt_polling( card );
      poll_wait( file, &card->wait_write, wait );
  }

  return mask;
}

#endif


static int dt_open( struct inode *inode, struct file *file ) 
{
  int minor = MINOR(inode->i_rdev);

  dprintk( "dt_open: Minor = %d, flags = %d.\n", minor, file->f_flags );

#ifdef LOG_DATA
  if( minor == DT_LOG_MINOR );  /* skip the following "if" */
  else  /* ... */
#endif

  if( minor >= NR_CARDS || !cards[minor].iobase ) return( -ENODEV );
  MOD_INC_USE_COUNT;
  return 0;
}


#ifdef KERNEL_2_0_x
static void dt_release( struct inode *inode, struct file *file )
#else
static int dt_release( struct inode *inode, struct file *file )
#endif
{
  struct dt_card *card = &cards[ MINOR(inode->i_rdev)];
  dprintk( "dt_release: minor = %d\n", MINOR(inode->i_rdev) );
  /* force speech of not yet processed data in the card */
  if( !card->in_use && 
      (card->flags & (DT_HAS_KERNEL|DT_IS_RUNNING|DT_HASDICT)) ==
        (DT_HAS_KERNEL|DT_IS_RUNNING|DT_HASDICT) ) 
    dt_sendbuff( card, "\xb", 1 );
  MOD_DEC_USE_COUNT;
#ifndef KERNEL_2_0_x
  return 0;
#endif
}


static int dt_ioctl( struct inode *inode, struct file *file,
                     unsigned int cmd, unsigned long arg )
{
  struct dt_card *card = &cards[MINOR(inode->i_rdev)];
  u_short iobase = card->iobase;
  int csval, ipval; 	/* used to load start addresses into the card */
  char *u_start; 	/* pointer to buffer in user space */
  unsigned int len;

  switch( cmd ){
    case DTIOLDKERNEL: /* request to load kernel to card */
      dprintk( "dt_ioctl: DTIOLDKERNEL\n" );
      {
        struct dt_kerninfo *dt_data = (struct dt_kerninfo *)arg;
        int retval = verify_area( VERIFY_READ, (void *)arg, 
                                  sizeof(struct dt_kerninfo) );
        if( retval ) return( retval );

	Get_user( u_start, &dt_data->dt_start );
        Get_user( len, &dt_data->dt_len ); /* length of kernel buffer */
        Get_user( csval, &dt_data->dt_csval );
        Get_user( ipval, &dt_data->dt_ipval );
        if( (retval = verify_area( VERIFY_READ, u_start, len )) )
          return( retval );
        dprintk( "dtpc: kernel is %d bytes, CS:IP = %04X:%04X\n", 
                 len, csval, ipval );

        /* initialize the card */
	card->flags = 0;
        dt_sendcmd( iobase, MODULE_reset ); /* reset the card */
        dt_waitmode( iobase, MODULE_init ); /* wait for card to initialize */
        dt_sendcmd( iobase, CMD_reset );
        if( dt_getstatus( iobase ) == MODULE_self_test ){
          if( ! dt_waitmode( iobase, MODULE_init ) ) return( -EIO );
        }
        dt_sendcmd( iobase, CMD_reset );
        if( ! dt_waitmode( iobase, MODE_status ) ) return( -EIO );
        dt_sendcmd( iobase, CMD_sync );
        if( ! dt_waitmode( iobase, MODE_error ) ) return( -EIO );
        dt_sendcmd( iobase, CMD_reset );
        if( ! dt_waitmode( iobase, MODE_status ) ) return( -EIO );
        dprintk( "dtpc: the card is ready\n" );

        /* the card is ready.  let's download its kernel. */
        dt_sendcmd( iobase, CMD_dma );
        if( ! dt_waitbit( iobase, STAT_dma_ready ) ) return( -EIO );
        if( dt_loadmem( card, 0x00000400, len, u_start ) ) return( -EIO );
        dt_delay( 100 );
        dprintk( "dtpc: the kernel is loaded\n" );

        /* the kernel is loaded, start it */
        dt_sendcmd( iobase, CMD_reset );
        if( ! dt_waitmode( iobase, MODE_status ) ) return( -EIO );
        dt_sendcmd( iobase, CMD_dma+1 );   /**xxx**/
	dt_delay( 100 );
        if( ! dt_waitbit( iobase, STAT_dma_ready ) ) return( -EIO );
        outb_p( DMA_control, iobase+4 );
        outb_p( DT_START_TASK, iobase+4 );
	dt_delay( 100 );

        /* send start address */
        outb_p( ipval & 0xFF, iobase+4 );
        outb_p( (ipval>>8) & 0xFF, iobase+4 );
        outb_p( csval & 0xFF, iobase+4 );
        outb_p( (csval>>8) & 0xFF, iobase+4 );

        /* The kernel in the card should be running, let's make sure */
	/* Don't know why but next line doesn't work, even if DEC's source 
	 * code does it like that.  An I/O trace from dosemu showed that
	 * the DOS loader waits for a 0x1000 status pattern.
	 */
//	if( ! dt_waitmode( iobase, MODE_ready|READY_kernel ) ) 
	if( ! dt_waitmode( iobase, 0x1000 ) ) 
	  return( -EIO );

        /* make sure the card responds to commands */
        dt_sendcmd( iobase, CMD_sync );
        if( ! dt_waitbit( iobase, STAT_cmd_ready ) ) return( -EIO );
	dt_delay( 100 );
	card->dt_dma_state = dt_getstatus( iobase ) & STAT_dma_state;
        dprintk( "dtpc: kernel is running!\n" );

        /* the kernel in the card is running, let's return success. */
        card->flags |= DT_HAS_KERNEL|DT_IS_RUNNING;
        break;
      }

    /* allocate memory on the card */
    case DTIOALLOCMEM:
      dprintk( "dt_ioctl: DTIOALLOCMEM\n" );
      {
        int retval;
	unsigned long paras, addr;

	retval = verify_area( VERIFY_READ, (void *)arg, sizeof(long) );
        if( retval ) return( retval );
        retval = verify_area( VERIFY_WRITE, (void *)arg, sizeof(long) );
        if( retval ) return( retval );
	Get_user( paras, (long*)arg );
	put_user( 0, (long*)arg );

	/* see if we can allocate memory on the card */
	if( ! dt_wait_dma( card ) ) return -EIO;
	outb_p( DMA_control, iobase+4 );
	outb_p( DT_MEM_ALLOC, iobase+4 );
	card->dt_dma_state ^= STAT_dma_state;

	if( ! dt_wait_dma( card ) ) return -EIO;
	outb_p( paras & 0xFF, iobase+4 );
	outb_p( (paras>>8) & 0xFF, iobase+4 );
	card->dt_dma_state ^= STAT_dma_state;

	/* Now, get the address of the memory we've just reserved */
	if( ! dt_wait_dma( card ) ) return -EIO;
	addr = inb_p( iobase+4 );
	addr |= (inb_p( iobase+4 )<<8);
	addr += (inb_p( iobase+4 )<<4);
	addr += (inb_p( iobase+4 )<<12);
	card->dt_dma_state ^= STAT_dma_state;

	dprintk( "dt_allocmem: paras = %lu addr = 0x%lX\n", paras, addr );
        put_user( addr, (unsigned long *)arg );
        break;
      }

    /* load up some memory on the card */
    case DTIOLDMEM:
      dprintk( "dt_ioctl: DTIOLDMEM\n" );
      {
	unsigned long tmp_addr;
        struct dt_loadinfo *dt_data = (struct dt_loadinfo *)arg;
        int retval = verify_area( VERIFY_READ, (void *)arg, 
                                  sizeof(struct dt_loadinfo) );
        if( retval ) return( retval );

        Get_user( u_start, &dt_data->dtl_start ); /* user space buffer addr */
        Get_user( len, &dt_data->dtl_len ); /* length of user buffer */
        if( (retval = verify_area( VERIFY_READ, u_start, len )) ) 
          return( retval );

        /* make sure we can really do this */
        Get_user( tmp_addr, &dt_data->dtl_addr );
        if( ! tmp_addr ) return( -ENOMEM ); /* out of memory on card */

        if( dt_loadmem( card, tmp_addr, len, u_start ) )
	  return( -EIO );
        break;
      }

    /* set a dictionary */
    case DTIOSETDICT:
      dprintk( "dt_ioctl: DTIOSETDICT\n" );
      {
        unsigned long l;
        struct dt_dicinfo *dt_data = (struct dt_dicinfo *)arg;
        int retval = verify_area( VERIFY_READ, (void *)arg, 
                                  sizeof(struct dt_dicinfo) );
        if( retval ) return( retval );

        if( ! dt_wait_dma( card ) ) return( -EIO );
        outb_p( DMA_control, iobase+4 );
        outb_p( DT_SET_DIC, iobase+4 );
	card->dt_dma_state ^= STAT_dma_state;

        if( ! dt_wait_dma( card ) ) return( -EIO );
        Get_user( l, &dt_data->dtd_addr ); /* load dict addr */
	/* we need CS:IP so we shift the adress */
	l = ((l << 12) & 0xFFFF0000) + (l & 0x0000000F);
        outb_p( l & 0xFF, iobase+4 );
        outb_p( (l>>8) & 0xFF, iobase+4 );
        outb_p( (l>>16) & 0xFF, iobase+4 );
        outb_p( (l>>24) & 0xFF, iobase+4 );

        Get_user( l, &dt_data->dtd_entries );
        outb_p( l & 0xFF, iobase+4 );
        outb_p( (l>>8) & 0xFF, iobase+4 );
        outb_p( (l>>16) & 0xFF, iobase+4 );
        outb_p( (l>>24) & 0xFF, iobase+4 );

	Get_user( l, &dt_data->dtd_type );
        outb_p( l & 0xFF, iobase+4 );
        outb_p( (l>>8) & 0xFF, iobase+4 );
	card->dt_dma_state ^= STAT_dma_state;

        Get_user( l, &dt_data->dtd_addr ); dprintk("dt_setdict: addr = %lX,", l );
        Get_user( l, &dt_data->dtd_entries ); dprintk(" entries = %lu,", l );
	Get_user( l, &dt_data->dtd_type ); dprintk(" type = %lu\n", l );

        /* the dictionary is set, we're done */
        card->flags |= DT_HASDICT;
        break;
      }

    case DTIOSTARTEXE:
      dprintk( "dt_ioctl: DTIOSTARTEXE\n" );
      {
        struct dt_taskinfo *dt_data = (struct dt_taskinfo *)arg;
        int retval = verify_area( VERIFY_READ, (void *)arg, 
                                  sizeof(struct dt_taskinfo) );
        if( retval ) return( retval );

        if( ! dt_wait_dma( card ) ) return( -EIO );
        outb_p( DMA_control, iobase+4 );
        outb_p( DT_START_TASK, iobase+4 );
	card->dt_dma_state ^= STAT_dma_state;

        /* send start address */
        Get_user( ipval, &dt_data->dts_ipval );
        Get_user( csval, &dt_data->dts_csval );
        if( ! dt_wait_dma( card ) ) return( -EIO );
        outb_p( ipval & 0xFF, iobase+4 );
        outb_p( (ipval>>8) & 0xFF, iobase+4 );
        outb_p( csval & 0xFF, iobase+4 );
        outb_p( (csval>>8) & 0xFF, iobase+4 );
	card->dt_dma_state ^= STAT_dma_state;

	dprintk( "dt_starttask: CS:IP = %04X:%04X\n", csval, ipval );
        break;
      }

    case DTIOTEST: /* This one is for testing, and not strictly necessary */
      dprintk( "dt_ioctl: DTIOTEST\n" );
      {
        dt_sendbuff( card, "Dectalk PC for Linux is running!\011", 34 );
        break;
      }

    case DTIOWARM:
      dprintk( "dt_ioctl: DTIOWARM\n" );
      {
	if( card->flags & (DT_HAS_KERNEL|DT_IS_RUNNING)) return( -EINVAL );
	break;
      }

    /* 
     * These are some direct speech control ioctls.
     */

    case DTFLUSH: 	/* flushes speech */
      dprintk( "dt_ioctl: DTFLUSH\n" );
      {
	int res;
	if( dt_excl( card ) ) return -ERESTARTSYS;
	res = dt_flush( card );
	card->in_use--;
	return res;
      }

    case DTFLUSHCHAR: 	/* flushes speech and say character */
      dprintk( "dt_ioctl: DTFLUSHCHAR\n" );
      {
	int res;
	if( dt_excl( card ) ) return -ERESTARTSYS;
	res = dt_flush_char( card, arg );
	card->in_use--;
	return res;
      }

    case DTVOLUP: 	/* augment volume by 5 */
      dprintk( "dt_ioctl: DTVOLUP\n" );
      {
	int res;
	if( dt_excl( card ) ) return -ERESTARTSYS;
	res = dt_ctrl( iobase, CTRL_vol_up, (arg)?arg:5 );
	card->in_use--;
	return res;
      }

    case DTVOLDOWN: 	/* reduce the volume by 5 */
      dprintk( "dt_ioctl: DTVOLDOWN\n" );
      {
	int res;
	if( dt_excl( card ) ) return -ERESTARTSYS;
	res = dt_ctrl( iobase, CTRL_vol_down, (arg)?arg:5 );
	card->in_use--;
	return res;
      }

    case DTVOLSET: 	/* set the volume level */
      dprintk( "dt_ioctl: DTVOLDOWN\n" );
      {
	int res;
	if( dt_excl( card ) ) return -ERESTARTSYS;
	if( arg > 100 ) arg = 100;
	if( arg < 0 ) arg = 0;
	res = dt_ctrl( iobase, CTRL_vol_set, arg );
	card->in_use--;
	return res;
      }

    case DTPAUSE: 	/* pause speech */
      dprintk( "dt_ioctl: DTPAUSE\n" );
      {
	int res;
	if( dt_excl( card ) ) return -ERESTARTSYS;
	res = dt_ctrl( iobase, CTRL_pause, 0 );
	card->in_use--;
	return res;
      }

    case DTRESUME: 	/* resume speech */
      dprintk( "dt_ioctl: DTRESUME\n" );
      {
	int res;
	if( dt_excl( card ) ) return -ERESTARTSYS;
	res = dt_ctrl( iobase, CTRL_resume, 0 );
	card->in_use--;
	return res;
      }


    /*
     * The following ioctls are included so programs expecting a real serial 
     * port are happy.
     */
    case TCGETS:
      dprintk( "dt_ioctl: TCGETS\n" );
      {
        int retval;
        retval = verify_area( VERIFY_WRITE, (void *)arg, 
                              sizeof( struct termios ) );
        if( retval ) return( retval );
        copy_to_user( (struct termios *)arg, &card->tios, 
                     sizeof( struct termios ) );
        break;
      }

    case TCSETSF:
    case TCSETSW:
    case TCSETS:
      dprintk( "dt_ioctl: TCSETS\n" );
      {
        int retval;
        retval = verify_area( VERIFY_READ, (void *)arg, 
                              sizeof( struct termios ) );
        if( retval ) return( retval );
        copy_from_user( &card->tios, (struct termios *)arg,
                       sizeof( struct termios ) );
        break;
      }

    case TCGETA:
      dprintk( "dt_ioctl: TCGETA\n" );
      {
#ifdef KERNEL_2_0_x
        int retval;
	struct termio tmp_termio;
        retval = verify_area( VERIFY_WRITE, (void*)arg, sizeof( struct termio ) );
        if( retval ) return retval;
	trans_to_termio( &card->tios, &tmp_termio );
	memcpy_tofs( (struct termio *)arg, &tmp_termio,
	    sizeof (struct termio ) );
#else
	if (kernel_termios_to_user_termio( (struct termio *)arg, &card->tios ) )
	    return -EFAULT;
#endif
	break;
      }

    default:
        dprintk( "dt_ioctl: default reached: 0x%X\n", cmd );
        return( -ENOIOCTLCMD );
  }
  return( 0 );
}



static struct file_operations dt_fops = {
#ifdef KERNEL_2_0_x
  NULL,  	/* lseek */
  dt_read,	/* read */
  dt_write,	/* write */
  NULL,		/* readdir */
  dt_select,	/* select */
  dt_ioctl,	/* ioctl */
  NULL,		/* mmap */
  dt_open,	/* open */
  dt_release	/* release */
#else
  NULL,  	/* lseek */
  dt_read,	/* read */
  dt_write,	/* write */
  NULL,		/* readdir */
  dt_poll,	/* poll */
  dt_ioctl,	/* ioctl */
  NULL,		/* mmap */
  dt_open,	/* open */
  NULL,		/* flush */
  dt_release	/* release */
#endif
};



#ifdef MODULE

int init_module (void)
{
  int status, found, i, num = 0;

  printk( Hello );
  dprintk( "dtpc: Start probing for DECtalk PC cards\n" );
  if( (dt_major = register_chrdev( DT_MAJOR, "dtpc", &dt_fops )) < 0 ){
    printk( "dtpc: unable to get major %d\n", DT_MAJOR );
    return -EIO;
  }

  for( i = 0; i < NR_CARDS; i++ ){
    dprintk( "dtpc: probing for DECtalk card at 0x%X\n", base[i] );
    found = 0;
    memset( &cards[i], 0, sizeof(struct dt_card) );
    if( check_region( base[i], 16 ) ) continue;  /* region not available */
    request_region( base[i], 16, "dtpc" );

    /* See if the card appears, without any prodding. */
    status = dt_getstatus( base[i] );
    dprintk( "dtpc: status = 0x%X\n", status );
    switch( status ){
      case MODULE_self_test:
      case MODULE_init:
        cards[num].status = status;
        found++;
        break;
    }

    if( ! found ){
      /* try to reset the card */
      dprintk( "dtpc: send MODULE_reset\n" );
      dt_sendcmd( base[i], MODULE_reset );
      /* wait for card to become ready */
      dt_delay( 1000 );
      status = dt_getstatus( base[i] );
      dprintk( "dtpc: status = 0x%X\n", status );
      switch( status ){
        case MODULE_self_test:
        case MODULE_init:
          cards[num].status = status;
          found++;
          break;
      }
    }
    
    if( found ){
      printk( "dtpc%d: DECtalk PC found at 0x%X\n", num, base[i] );
      cards[num].iobase = base[i];
      init_timer( &cards[num++].polling_timer );
    }else{
      /* no card for this region */
      release_region( base[i], 16 );
    }
  }

  if( num ) return( 0 );
  else{
    printk( "dtpc: No DECtalk board found\n" );
    unregister_chrdev( dt_major, "dtpc" );
    return( -ENODEV );
  }
}


void cleanup_module(void)
{
  int i;

  if( MOD_IN_USE ) {
    printk( "dtpc: module is still in use...\n" );
    /* return; */
  }

  unregister_chrdev( dt_major, "dtpc" );
  for( i = 0; i < NR_CARDS; i++ ){
    del_timer( &cards[i].polling_timer );
    if( cards[i].iobase ) release_region( cards[i].iobase, 16 );
  }
  dprintk( "dtpc: done cleanup\n" );
}

#else
#warning This driver is not configured to be compiled in the kernel.....
#warning For instance the initialization is never called....
#endif

